/*
Copyright  2007 Eland Systems All Rights Reserved.

   1. Redistribution and use in source and binary forms must retain the above
   copyright notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this software
	must display the following acknowledgement:
         This product includes software developed by Eland Systems.

   4. The name of Eland Systems may not be used to endorse or promote products
   derived from this software without specific prior written permission.

   Author scam+protect@elandsys.com
*/

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sysexits.h>
#include <unistd.h>
#include <errno.h>
#include <stdarg.h>
#include <ctype.h>
#include <pwd.h>
#include <grp.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <syslog.h>

#include <pthread.h>

#include "libmilter/mfapi.h"

#ifndef LINUX
#include <sys/queue.h>
#else
#include "queue.h"
#endif

#include "lock.h"
#include "util.h"

#define PIDFILE "/var/spool/scam/scam-protect.pid"
#define MLFIPRIV	((struct mlfiPriv *) smfi_getpriv(ctx))

#define VERSION "1.0"

/* POPAUTH db */
#ifdef POPAUTH
# define POPAUTHDB	"/etc/mail/popip.db"
#endif

char *daemonmta = NULL;
char *contact = NULL;

struct mlfiPriv {
	char	*queueid;
	int	msa;
	int	smtpauth;
} mlfi_priv;


static sfsistat mlfi_cleanup(SMFICTX *);

static sfsistat
mlfi_connect(SMFICTX *ctx, char *hostname,  _SOCK_ADDR *hostaddr)
{
	struct mlfiPriv *priv= MLFIPRIV;
	struct sockaddr_in  *remoteaddr;
	char *daemonname;

	remoteaddr = (struct sockaddr_in *) hostaddr;

	if ((remoteaddr != NULL) && (remoteaddr->sin_family == AF_INET))
	{
		if (remoteaddr->sin_addr.s_addr == inet_addr("127.0.0.1"))
			return SMFIS_ACCEPT;
	}

	if (check_ip(remoteaddr->sin_addr.s_addr) == 1)
	{
		return SMFIS_ACCEPT;
	}

	priv = malloc(sizeof *priv);
	if (priv == NULL) {
		syslog (LOG_ERR, "cannot allocate priv");
		return SMFIS_TEMPFAIL;
	}
	memset(priv, '\0', sizeof *priv);

	if (smfi_setpriv(ctx, priv) != MI_SUCCESS)
	{
		smfi_setpriv(ctx, priv);
		free(priv);

		syslog (LOG_ERR, "cannot setpriv");
		(void) mlfi_cleanup(ctx);
        return SMFIS_TEMPFAIL;
	}

	priv->queueid = NULL;

	daemonname = smfi_getsymval(ctx, "{daemon_name}");
	if (daemonname != NULL)
	{
		if (strcasecmp(daemonname, "MSA") == 0)
		{
				priv->msa = 1;
		} else {

			if (daemonmta != NULL)
			{
				if (strcasecmp(daemonname, daemonmta) == 0)
				{
					priv->msa = 1;
				}
			}
#if POPAUTH
			if (priv->msa == 0)
				priv->msa = checkpopauth(POPAUTHDB, hostaddr);
#endif
		}

	}

	if (priv->msa == 1)
	{
		return SMFIS_ACCEPT;
	}

	return SMFIS_CONTINUE;
}

static sfsistat
mlfi_envfrom(SMFICTX *ctx, char **argv)
{
	struct mlfiPriv *priv= MLFIPRIV;
	char *authtype;
	char *queueid;
	char *mailaddr;
	char *domain = NULL;

	authtype = smfi_getsymval(ctx, "{auth_type}");

	if (authtype != NULL)
	{
		return SMFIS_ACCEPT;
	}

	queueid= smfi_getsymval(ctx, "i");
	if (queueid != NULL)
	{
		if (priv->queueid != NULL)
			free(priv->queueid);

		if ((priv->queueid = strdup (queueid)) == NULL)
		{
			syslog (LOG_ERR, "cannot allocate queueid");
			(void) mlfi_cleanup(ctx);
			return SMFIS_TEMPFAIL;
		}
	}

	mailaddr= parseaddr( *argv);
	if (mailaddr != NULL)
		domain = rfc822domain(mailaddr);

	if ((domain != NULL) && (check_dom( domain) == 1))
	{
		smfi_setreply( ctx, "550", "5.7.0", "Restricted email address");
		return SMFIS_REJECT;
	}

	return SMFIS_CONTINUE;
}

static sfsistat
mlfi_header(SMFICTX *ctx, char *headerf, char *headerv)
{

	struct mlfiPriv *priv= MLFIPRIV;
	char *mailaddr = NULL;
	char *domain = NULL;

	if ((headerf != NULL)  && (strcasecmp(skipws(headerf), "From") == 0))
	{
		mailaddr= parseaddr( headerv);

		if (mailaddr != NULL)
		{
			domain = rfc822domain(mailaddr);
			if ((domain != NULL) && (check_dom( domain) == 1))
			{
				syslog(LOG_INFO, "%s: Header From %s", priv->queueid, mailaddr);
				free(mailaddr);
				smfi_setreply( ctx, "550", "5.7.0", "Message rejected - restricted email address");
				return SMFIS_REJECT;
			}
			free(mailaddr);
		}
	}

	return SMFIS_CONTINUE;
}

sfsistat
mlfi_abort(SMFICTX *ctx)
{

	mlfi_cleanup(ctx);

	return SMFIS_CONTINUE;
}

static sfsistat
mlfi_cleanup(SMFICTX *ctx)
{
	struct mlfiPriv *priv = MLFIPRIV;

	if (priv == NULL)
		return SMFIS_CONTINUE;

	if (priv->queueid)
	{
		free(priv->queueid);
		priv->queueid = NULL;
	}

	return SMFIS_CONTINUE;

}

static sfsistat
mlfi_close(SMFICTX *ctx)
{
	struct mlfiPriv *priv = MLFIPRIV;

	mlfi_cleanup(ctx);

	if (priv)
	{
		free(priv);
	}
	smfi_setpriv(ctx, NULL);

	return SMFIS_CONTINUE;
}

struct smfiDesc smfilter =
{
	"scam-protect",
	SMFI_VERSION,
	SMFIF_ADDHDRS,
	mlfi_connect,
	NULL,
	mlfi_envfrom,
	NULL,
	mlfi_header,
	NULL,
	NULL,
	NULL,
	mlfi_abort,
	mlfi_close,
};

static void
usage()
{
	fprintf(stdout, "scam-protect version %s\n", VERSION);
    fprintf(stdout, "usage: scam-protect -p protocol:address [-m daemonname] [-u user] [-g group] [-D]\n");
	fprintf(stdout, "          -C append message to SMTP reply\n");
	fprintf(stdout, "          -D run as a daemon\n");
	fprintf(stdout, "          -u run as specified user\n");
	fprintf(stdout, "          -g run as specified group\n");
}

static void
catch_signal( int signo )
{
	switch (signo)
	{
		case SIGUSR1:
		   syslog (LOG_NOTICE, "Caught SIGUSR1, exiting...");
			smfi_stop();
			sleep(10);
			exit (0) ;
	}

}

int
write_pid()
{
	char buf[20];
	int fd;
	long pid;

	if ((fd = open (PIDFILE, O_CREAT | O_TRUNC | O_WRONLY, 0600)) == -1)
	{
		syslog (LOG_ERR, "cannot open pidfile ");
	} else {
		pid = getpid ();
		snprintf (buf, sizeof (buf), "%ld", (long) pid);
		if (write (fd, buf, strlen (buf)) != strlen (buf)) {
			syslog (LOG_INFO, "cannot write to pidfile ");
		}
		close(fd);
	}
	return pid;
}

void
daemonize()
{

	signal (SIGUSR1, catch_signal);

	(void)close(0);
	(void)open("/dev/null", O_RDONLY, 0);
	(void)close(1);
	(void)open("/dev/null", O_RDONLY, 0);
	(void)close(2);
	(void)open("/dev/null", O_RDONLY, 0);

	switch (fork ())
	{
		case 0:
			syslog (LOG_INFO, "scam-protect running %s", VERSION);
			break;

		case -1:
			perror ("cannot fork\n");
			exit (3);
			break;

		default:
			exit(0);
	}

	if (setsid() < 0)
		exit(1);

	write_pid();
	chdir("/");

}

int
main(int argc, char *argv[])
{
	int c;
	int ret;
	const char *args = "p:T:h:u:g:m:C:d:D";
	extern char *optarg;
	int daemonmode = 0;
	struct passwd *passwd = NULL;
	struct group *grp;
	char *user = NULL;
	char *group = NULL;
	uid_t uid;
	gid_t gid;

	umask(077);

	/* already a daemon */
	if(getppid()==1)
		return 0;

	/* Process command line options */
	while ((c = getopt(argc, argv, args)) != (char)EOF)
	{
		switch (c)
		{
		  case 'p':
			if (optarg == NULL || *optarg == '\0')
			{
				(void) fprintf(stderr, "Illegal connection: %s\n", optarg);
				exit(EX_USAGE);
			}
			if (smfi_setconn(optarg) == MI_FAILURE)
			{
				(void) fprintf(stderr, "smfi_setconn failed");
				exit(EX_SOFTWARE);
			}

			/*
				** If we're using a local socket, make sure it doesn't
				** already exist.
			*/
            if(strncmp(optarg, "unix:", 5) == 0)
                unlink(optarg + 5);
            else if(strncmp(optarg, "local:", 6) == 0)
                unlink(optarg + 6);

			break;

		  case 'T':
            if (optarg == NULL || *optarg == '\0')
            {
                (void) fprintf(stderr, "Illegal timeout: %s\n", optarg);
                exit(EX_USAGE);
            }
            if(smfi_settimeout(atoi(optarg)) == MI_FAILURE)
            {
                (void) fputs("smfi_settimeout failed", stderr);
                exit(EX_SOFTWARE);
            }
            break;

			case 'u':
				if (optarg == NULL || *optarg == '\0')
				{
					(void) fprintf(stderr, "Invalid username: %s\n", optarg);
					exit(EX_USAGE);
				}
				user = strdup(optarg);
				if ((passwd = getpwnam(user)) != NULL)
				{
					uid = passwd->pw_uid;
					gid = passwd->pw_gid;
				}
				break;

			case 'g':
				if (optarg == NULL || *optarg == '\0')
				{
					(void) fprintf(stderr, "Invalid group name: %s\n", optarg);
					exit(EX_USAGE);
				}
				group = strdup(optarg);
				if ((grp = getgrnam(group)) != NULL)
				{
					gid = grp->gr_gid;
				}
				break;

		  case 'd':
			if (optarg == NULL || *optarg == '\0')
            {
                (void) fprintf(stderr, "Illegal debug value: %s\n", optarg);
                exit(EX_USAGE);
            }
			if (smfi_setdbg(atoi(optarg)) == MI_FAILURE)
				fprintf(stderr, "smfi_setdbg enabled\n");
			break;

		  case 'D':
			  daemonmode = 1;
			  break;

		  case 'm':
				if (optarg == NULL || *optarg == '\0')
				{
					(void) fprintf(stderr, "Invalid daemonname: %s\n", optarg);
					exit(EX_USAGE);
				}
				daemonmta = strdup(optarg);
				break;

		case 'C':
				if (optarg == NULL || *optarg == '\0')
				{
					(void) fprintf(stderr, "Invalid contact: %s\n", optarg);
					exit(EX_USAGE);
				}
				if (strlen(optarg) > 100)
				{
					(void) fprintf(stderr, "Contact should be less than 100 characters: %s\n", optarg);
					exit(EX_USAGE);
				}
				contact = strdup(optarg);
				break;


		  case 'h':
		  default:
			usage();
			exit(0);

		}
	}

	(void) closelog();
	openlog("scam-protect", LOG_PID | LOG_NDELAY, LOG_MAIL);

	setgid(gid);
	setuid(uid);

	if (!getuid() || !geteuid())
	{
		syslog (LOG_ERR, "scam-protect cannot run as root");
		exit(0);
	}

	if (smfi_register(smfilter) == MI_FAILURE)
	{
		fprintf(stderr, "smfi_register failed\n");
		syslog (LOG_ERR, "smfi_register failed");
		exit(EX_UNAVAILABLE);
	}


	white_init();
	protect_readconf();

	if (daemonmode == 1)
	{
		daemonize();
	}

	ret = smfi_main();
	unlink(PIDFILE);
	syslog (LOG_INFO, "Exit");
	return(ret);
}

