/*
Copyright  2004-2007 Eland Systems All Rights Reserved.

   1. Redistribution and use in source and binary forms must retain the above
   copyright notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this software
	must display the following acknowledgement:
         This product includes software developed by Eland Systems.

   4. The name of Eland Systems may not be used to endorse or promote products
   derived from this software without specific prior written permission.

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sysexits.h>
#include <syslog.h>
#include <pthread.h>

#include <netdb.h>
#include <arpa/nameser.h>
#include <resolv.h>

#ifndef LINUX
#include <sys/queue.h>
#else
#include "queue.h"
#endif

#include "lock.h"
#include "util.h"

#ifdef POPAUTH
# include <db.h>
# include <fcntl.h>
#endif

#include <regex.h>

#ifndef	REG_BASIC
#define	REG_BASIC 0
#endif

#define LINELENGTH 1024

#define SCAMCONF "/etc/mail/scam.conf"

static unsigned char lowercase[256] = {
    0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15,
   16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31,
  ' ','!','"','#','$','%','&', 39,'(',')','*','+',',','-','.','/',
  '0','1','2','3','4','5','6','7','8','9',':',';','<','=','>','?',
  '@','a','b','c','d','e','f','g','h','i','j','k','l','m','n','o',
  'p','q','r','s','t','u','v','w','x','y','z','[', 92,']','^','_',
  '`','a','b','c','d','e','f','g','h','i','j','k','l','m','n','o',
  'p','q','r','s','t','u','v','w','x','y','z','{','|','}','~',127,
  128,129,130,131,132,133,134,135,136,137,138,139,140,141,142,143,
  144,145,146,147,148,149,150,151,152,153,154,155,156,157,158,159,
  160,161,162,163,164,165,166,167,168,169,170,171,172,173,174,175,
  176,177,178,179,180,181,182,183,184,185,186,187,188,189,190,191,
  192,193,194,195,196,197,198,199,200,201,202,203,204,205,206,207,
  208,209,210,211,212,213,214,215,216,217,218,219,220,221,222,223,
  224,225,226,227,228,229,230,231,232,233,234,235,236,237,238,239,
  240,241,242,243,244,245,246,247,248,249,250,251,252,253,254,255};

#define LPC(x) (lowercase[(unsigned char)(x)])

pthread_rwlock_t white_lock;

struct CIDR {
	unsigned long ip;
	short int masklen;
	SLIST_ENTRY(CIDR)	cidrs;
};

SLIST_HEAD(, CIDR)	cidrlist;

struct doment {
	char	*domain;
	SLIST_ENTRY(doment)	domentries;
};

SLIST_HEAD(, doment) domlist;

size_t
strlcpy(char *dst, const char *src, size_t siz)
{
	register char *d = dst;
	register const char *s = src;
	register size_t n = siz;

	/* Copy as many bytes as will fit */
	if (n != 0 && --n != 0) {
		do {
			if ((*d++ = *s++) == 0)
				break;
		} while (--n != 0);
	}

	/* Not enough room in dst, add NUL and traverse rest of src */
	if (n == 0) {
		if (siz != 0)
			*d = '\0';		/* NUL-terminate dst */
		while (*s++)
			;
	}

	return(s - src - 1);	/* count does not include NUL */
}

void white_init()
{
	int ret;

	if ((ret = pthread_rwlock_init(&white_lock, NULL)) != 0)
	{
		syslog(LOG_ERR,  "pthread_rwlock_init failed: %s", strerror(ret));
		exit(EX_OSERR);
	}
	SLIST_INIT(&cidrlist);
	SLIST_INIT(&domlist);
}

int protect_readconf()
{
	FILE *fh;
	char line[LINELENGTH + 1];
	char buf[LINELENGTH + 1];
	struct CIDR *entcidr;

	struct doment *domadd;
	int a;

	int lline;
	int lineno = 0;

	WRLOCK(white_lock);

	while (!SLIST_EMPTY(&cidrlist))
	{
		entcidr = SLIST_FIRST(&cidrlist);
		SLIST_REMOVE_HEAD(&cidrlist, cidrs);
		free(entcidr);
	}

	while (!SLIST_EMPTY(&domlist))
	{
		domadd = SLIST_FIRST(&domlist);
		if (domadd->domain != NULL)
			free(domadd->domain);
		SLIST_REMOVE_HEAD(&domlist, domentries);
		free(domadd);
	}

	if ((fh = fopen( SCAMCONF, "r")) == NULL)
	{
		UNLOCK(white_lock);
		return 1;
	}

	syslog (LOG_INFO, "reading conf");

	while( fgets(line, LINELENGTH, fh) )
	{
		lineno++;
		if (line[0] == '#')
			continue;

		lline = strlen(line);
		if (lline > 16)
		{
			unsigned int ip1 = 0;
			unsigned int ip2 = 0;
			unsigned int ip3 = 0;
			unsigned int ip4 = 0;
			unsigned int ipm = 0;
			int j = 0;

			while ((j < (lline -1)) && (line[j] > 64))
			{
				line[j] = LPC(line[j]);
				j++;
			}

			if (sscanf( line, "protectallow:%u.%u.%u.%u/%u", &ip1, &ip2, &ip3, &ip4, &ipm) == 5)
			{
				snprintf( buf, sizeof(buf), "%u.%u.%u.%u", ip1, ip2, ip3, ip4);
				entcidr = (struct CIDR *)malloc(sizeof(struct CIDR));
				entcidr->ip = inet_addr(buf);
				entcidr->masklen = ipm;
				SLIST_INSERT_HEAD(&cidrlist, entcidr, cidrs);
				syslog (LOG_DEBUG, "ProtectAllow set %s/%d", buf, ipm);
			}
			else if (sscanf( line, "protectallow:%u.%u.%u.%u", &ip1, &ip2, &ip3, &ip4) == 4)
			{
				snprintf( buf, sizeof(buf), "%u.%u.%u.%u", ip1, ip2, ip3, ip4);
				entcidr = (struct CIDR *)malloc(sizeof(struct CIDR));
				entcidr->ip = inet_addr(buf);
				entcidr->masklen = 32;
				SLIST_INSERT_HEAD(&cidrlist, entcidr, cidrs);
				syslog (LOG_DEBUG, "ProtectAllow set %s", buf);
			}
			else if (sscanf( line, "protectdomain:%256[A-Za-z0-9-.]", buf) == 1)
			{
				if (strlen(buf) > 3)
				{
					domadd = (struct doment *)malloc(sizeof(struct doment));
					if ((domadd->domain = strdup(buf)) == NULL)
					{
						return 1;
					}
					SLIST_INSERT_HEAD(&domlist, domadd, domentries);
					syslog (LOG_DEBUG, "ProtectDomain %s", buf);
				}
			}
		}
	}
	fclose(fh);
	UNLOCK(white_lock);
	return 0;
}

int
check_dom(const char* domname)
{
	int ret = 0;
	struct doment *entre;
	char *domainp = NULL;
	int cmp;
	char *domnameLC = NULL;
	char *cp = NULL;
	char *entredomainLC = NULL;

	if ( strlen( domname ) < 1 )
		return ret;


	domnameLC = strdup( domname );
	if ( domnameLC == NULL )
		return ( ret );

	for ( cp = domnameLC; *cp; ++cp )
		*cp = tolower( *cp );

	SLIST_FOREACH( entre, &domlist, domentries )
	{
		if (entredomainLC != NULL)
			free(entredomainLC);

		entredomainLC = strdup( entre->domain );
		if ( entredomainLC == NULL )
		{
			free( domnameLC );
			return ret;
		}

		for ( cp = entredomainLC; *cp; ++cp )
			*cp = tolower( *cp );

		cmp = strcmp( domnameLC, entredomainLC );

		if ( cmp == 0 )
		{
			ret = 1;
			break;
		}

		if (cmp > 0)
		{
			domainp = strstr( domnameLC, entredomainLC );

			if ( domainp == NULL )
				continue;

			cmp = strcmp( domainp, entredomainLC );

			if (( cmp == 0) && (domainp != domnameLC) && (*(domainp - 1) == '.' ))
			{
				ret = 1;
				break;
			}
		}
	}

	free( domnameLC );
	if (entredomainLC != NULL)
		free( entredomainLC );

	return ret;
}

int
ip_cidr(const unsigned long ip, const short int masklen, const unsigned long checkip)
{
	unsigned long ipaddr = 0;
	unsigned long cidrip = 0;
	unsigned long subnet = 0;

	subnet = ~0;
	subnet = subnet << (32 - masklen);

	cidrip = htonl(ip) & subnet;
	ipaddr = ntohl(checkip) & subnet;

	if (cidrip == ipaddr)
		return 1;

	return 0;
}

int
check_ip(const unsigned long checkip)
{
	int ret = 0;
	struct CIDR *entcidr;

	RDLOCK(white_lock);
	SLIST_FOREACH(entcidr, &cidrlist, cidrs)
	{
		if (ip_cidr(entcidr->ip, entcidr->masklen, checkip) == 1)
		{
			syslog (LOG_INFO, "skip cidr");
			ret = 1;
			break;
		}
	}
	UNLOCK(white_lock);

	return ret;
}

char *skipws(char *cp)
{
	while (*cp == ' ')
			cp++;

	return cp;
}

char *findbalance(char *c, unsigned int cb)
{
	while (*c)
	{
		if (*c == cb)
			return c;
		c++;
	}
	return NULL;
}

char *parsespecial(char *c)
{
	while (*c)
	{
		switch (*c)
		{
			case '"':
				c = findbalance(c+1, '"');
				if (c == NULL)
					return c;
				break;

			case '(':
				c = findbalance(c+1, ')');
				if (c == NULL)
					return c;
				break;

			case '[':
				c = findbalance(c+1, ']');
				if (c == NULL)
					return c;
				break;

			case '\\':
				if ((*(c+1)) != '\0')
					c++;
				break;

			case '<':
			case ' ':
				return c;
				break;
		}
		c++;
	}

	return c;
}

char *parseaddr(char *c)
{
	char *s;
	char *a;
	int i;

	if ((a= malloc(1)) == NULL)
		return NULL;

	a[0] = '\0';

	while (c != NULL)
	{
		c = skipws(c);
		s = c;
		c = parsespecial(c);
		if (c - s != 0)
			i = c -s;
		else
			i = strlen(c);

		if ((a = realloc(a, i +1)) == NULL)
		{
			free(a);
			return NULL;
		}
		memcpy( a, s, i);
		a[i] = '\0';

		if ((a[0] != '"') && (strchr(a, '@') > 0))
		{
			if (a[0] == '<')
			{
				s = a + 1;
				c = findbalance(a+1, '>');
				if (c != NULL)
				{
					a = realloc(a, c -s + 1);
					memcpy( a, s, c -s);
					a[c - s] = '\0';
				}
			}
			return a;
		}
		if ( c -s <= 0)
			break;
	}
	free(a);
	return NULL;
}

char* rfc822domain( char* rfc822addr )
{
	char *p = rfc822addr;
	char *q;
	int k;

	k = strlen( p );

	if ( k >= 1 )
	{
		for ( q = p + k - 1; q != p; --q )
		{
			if ( *(q - 1) == '@' )
			{
				p = q;
				break;
			}
		}
	}

	return p;
}

#ifdef POPAUTH

int
checkpopauth(const char *dbfile, struct sockaddr *ip)
{
	int ret = 0;
	DB *db;
	struct sockaddr_in *sin;
	struct in_addr addr;
	DBT d;
	DBT q;
	char ipbuf[17];

	/* skip anything not IPv4 (for now) */
	if (ip->sa_family != AF_INET)
		return 0;
	else
		sin = (struct sockaddr_in *)ip;

	memset(&d, 0, sizeof d);
	memset(&q, 0, sizeof q);

	memcpy(&addr.s_addr, &sin->sin_addr, sizeof addr.s_addr);

	inet_ntop(AF_INET, &addr, ipbuf, sizeof ipbuf);

#ifdef POPTEXT
	q.data = sprintf("POP:%s",ipbuf);
#else
	q.data = ipbuf;
#endif
	q.size = strlen(q.data);
#if DB_VERSION_MAJOR > 3
	if (db_create(&db, NULL, 0) != 0 ||
	    db->open(db, NULL, dbfile, NULL, DB_HASH, DB_RDONLY, 0) != 0)
#elif DB_VERSION_MAJOR == 3
	if (db_create(&db, NULL, 0) != 0 ||
	    db->open(db, dbfile, NULL, DB_HASH, DB_RDONLY, 0) != 0)
#elif DB_VERSION_MAJOR == 2
	if (db_open(dbfile, DB_HASH, DB_RDONLY, DB_MODE, NULL, NULL, &db) != 0)
#else /* DB_VERSION_MAJOR == 2 */
	if ((db = db_open(dbfile, O_RDONLY, DB_MODE, DB_HASH, NULL)) == NULL)
#endif /* DB_VERSION_MAJOR < 2 */
		return 0;

#if DB_VERSION_MAJOR < 2
	if (!((db->get) (db, &q, &d, 0)))
#else /* DB_VERSION_MAJOR < 2 */
	if (!((db->get) (db, NULL, &q, &d, 0)))
#endif
		ret = 1;

#if DB_VERSION_MAJOR < 2
	(void) db->close(db);
#else /* DB_VERSION_MAJOR < 2 */
	(void) db->close(db, 0);
#endif

	return ret;
}
#endif

